@@ -21,10 +21,11 @@
 #include <stdarg.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <string.h>
+#include <avahi-common/malloc.h>
 
 #include "salut-avahi-errors.h"
 #include "salut-avahi-entry-group.h"
-#include "salut-avahi-entry-group-signals-marshal.h"
 #include "salut-avahi-entry-group-enumtypes.h"
 
 G_DEFINE_TYPE(SalutAvahiEntryGroup, salut_avahi_entry_group, G_TYPE_OBJECT)
@@ -50,14 +51,14 @@ typedef struct _SalutAvahiEntryGroupPrivate SalutAvahiEntryGroupPrivate;
 
 struct _SalutAvahiEntryGroupPrivate
 {
-  SalutAvahiEntryGroupState state; 
+  SalutAvahiEntryGroupState state;
   SalutAvahiClient *client;
   AvahiEntryGroup *group;
   GHashTable *services;
   gboolean dispose_has_run;
 };
 
-typedef struct _SalutAvahiEntryGroupServicePrivate 
+typedef struct _SalutAvahiEntryGroupServicePrivate
                 SalutAvahiEntryGroupServicePrivate;
 
 struct _SalutAvahiEntryGroupServicePrivate {
@@ -67,6 +68,15 @@ struct _SalutAvahiEntryGroupServicePrivate {
   GHashTable *entries;
 };
 
+typedef struct _SalutAvahiEntryGroupServiceEntry
+                SalutAvahiEntryGroupServiceEntry;
+
+struct _SalutAvahiEntryGroupServiceEntry {
+  guint8 *value;
+  gsize size;
+};
+
+
 #define SALUT_AVAHI_ENTRY_GROUP_GET_PRIVATE(o)     (G_TYPE_INSTANCE_GET_PRIVATE ((o), SALUT_TYPE_AVAHI_ENTRY_GROUP, SalutAvahiEntryGroupPrivate))
 
 static void
@@ -125,13 +135,13 @@ salut_avahi_entry_group_class_init (SalutAvahiEntryGroupClass *salut_avahi_entry
                                  G_PARAM_STATIC_BLURB);
   g_object_class_install_property(object_class, PROP_STATE, param_spec);
 
-  signals[STATE_CHANGED] = 
+  signals[STATE_CHANGED] =
     g_signal_new("state-changed",
                  G_OBJECT_CLASS_TYPE(salut_avahi_entry_group_class),
                  G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                  0,
                  NULL, NULL,
-                 salut_avahi_entry_group_marshal_VOID__INT,
+                 g_cclosure_marshal_VOID__ENUM,
                  G_TYPE_NONE, 1, SALUT_TYPE_AVAHI_ENTRY_GROUP_STATE);
 }
 
@@ -139,7 +149,7 @@ void
 salut_avahi_entry_group_dispose (GObject *object)
 {
   SalutAvahiEntryGroup *self = SALUT_AVAHI_ENTRY_GROUP (object);
-  SalutAvahiEntryGroupPrivate *priv = 
+  SalutAvahiEntryGroupPrivate *priv =
     SALUT_AVAHI_ENTRY_GROUP_GET_PRIVATE (self);
 
   if (priv->dispose_has_run)
@@ -174,10 +184,10 @@ salut_avahi_entry_group_finalize (GObject *object)
   G_OBJECT_CLASS (salut_avahi_entry_group_parent_class)->finalize (object);
 }
 
-static 
+static
 void _free_service (gpointer data) {
   SalutAvahiEntryGroupService *s = (SalutAvahiEntryGroupService *)data;
-  SalutAvahiEntryGroupServicePrivate *p = 
+  SalutAvahiEntryGroupServicePrivate *p =
     (SalutAvahiEntryGroupServicePrivate *)s;
   g_free(s->name);
   g_free(s->type);
@@ -189,7 +199,7 @@ void _free_service (gpointer data) {
 
 static GQuark
 detail_for_state(AvahiEntryGroupState state) {
-  static struct { AvahiClientState state; gchar *name; GQuark quark; } 
+  static struct { AvahiClientState state; gchar *name; GQuark quark; }
     states[]  = { { AVAHI_ENTRY_GROUP_UNCOMMITED,  "uncommited", 0 },
                   { AVAHI_ENTRY_GROUP_REGISTERING, "registering", 0 },
                   { AVAHI_ENTRY_GROUP_ESTABLISHED, "established", 0 },
@@ -211,10 +221,10 @@ detail_for_state(AvahiEntryGroupState state) {
 }
 
 static void
-_avahi_entry_group_cb(AvahiEntryGroup *g, 
+_avahi_entry_group_cb(AvahiEntryGroup *g,
                       AvahiEntryGroupState state, void *data) {
   SalutAvahiEntryGroup *self = SALUT_AVAHI_ENTRY_GROUP(data);
-  SalutAvahiEntryGroupPrivate *priv = 
+  SalutAvahiEntryGroupPrivate *priv =
     SALUT_AVAHI_ENTRY_GROUP_GET_PRIVATE (self);
 
   /* Avahi can call the callback before return from _client_new */
@@ -223,7 +233,7 @@ _avahi_entry_group_cb(AvahiEntryGroup *g,
 
   g_assert(g == priv->group);
   priv->state = state;
-  g_signal_emit(self, signals[STATE_CHANGED], 
+  g_signal_emit(self, signals[STATE_CHANGED],
                 detail_for_state(state),  state);
 }
 
@@ -232,24 +242,99 @@ salut_avahi_entry_group_new(void) {
   return g_object_new(SALUT_TYPE_AVAHI_ENTRY_GROUP, NULL);
 }
 
+static guint
+_entry_hash(gconstpointer v) {
+  const SalutAvahiEntryGroupServiceEntry *entry =
+      (const SalutAvahiEntryGroupServiceEntry *)v;
+  guint32 h = 0;
+  gint i;
+
+  for (i = 0; i < entry->size; i++) {
+    h = (h << 5) - h + entry->value[i];
+  }
+
+  return h;
+}
+
+static gboolean
+_entry_equal(gconstpointer a, gconstpointer b) {
+  const SalutAvahiEntryGroupServiceEntry *aentry =
+      (const SalutAvahiEntryGroupServiceEntry *)a;
+  const SalutAvahiEntryGroupServiceEntry *bentry =
+      (const SalutAvahiEntryGroupServiceEntry *)b;
+
+  if (aentry->size != bentry->size) {
+    return FALSE;
+  }
+
+  return memcmp(aentry->value, bentry->value, aentry->size) == 0;
+}
+
+static SalutAvahiEntryGroupServiceEntry *
+_new_entry(const guint8 *value, gsize size) {
+  SalutAvahiEntryGroupServiceEntry *entry;
+
+  if (value == NULL) {
+    return NULL;
+  }
+
+  entry = g_slice_new(SalutAvahiEntryGroupServiceEntry);
+  entry->value = g_malloc (size + 1);
+  memcpy (entry->value, value, size);
+  /* for string keys, make sure it's NUL-terminated too */
+  entry->value[size] = 0;
+  entry->size = size;
+
+  return entry;
+}
+
+static void
+_set_entry(GHashTable *table, const guint8 *key, gsize ksize,
+    const guint8 *value, gsize vsize) {
+
+  g_hash_table_insert(table, _new_entry(key, ksize), _new_entry(value, vsize));
+}
+
+static void
+_free_entry(gpointer data) {
+  SalutAvahiEntryGroupServiceEntry *entry =
+      (SalutAvahiEntryGroupServiceEntry *)data;
+
+  if (entry == NULL) {
+    return;
+  }
+
+  g_free(entry->value);
+  g_slice_free(SalutAvahiEntryGroupServiceEntry, entry);
+}
+
 static GHashTable *
 _string_list_to_hash(AvahiStringList *list) {
   GHashTable *ret;
-  ret = g_hash_table_new_full(g_str_hash,
-                              g_str_equal,
-                              g_free,
-                              g_free);
+  ret = g_hash_table_new_full(_entry_hash,
+                              _entry_equal,
+                              _free_entry,
+                              _free_entry);
   AvahiStringList *t;
-  for (t = list ; t != NULL; t = t->next) {
+  for (t = list ; t != NULL; t = avahi_string_list_get_next(t)) {
+    gchar *key;
     gchar *value;
-    value = g_strstr_len((gchar *)t->text, t->size, "=");
+    gsize size;
+    int r;
+
+    /* list_get_pair only fails if if memory allocation fails. Normal glib
+     * behaviour is to assert/abort when that happens */
+    r = avahi_string_list_get_pair(t, &key, &value, &size);
+    g_assert(r == 0);
+
     if (value == NULL) {
-      g_hash_table_insert(ret, g_strndup((gchar *)t->text, t->size), NULL);
+      _set_entry(ret, t->text, t->size, NULL, 0);
     } else {
-      int offset = value - (gchar *)t->text;
-      g_hash_table_insert(ret, g_strndup((gchar *)t->text, offset),
-                               g_strndup(value + 1, t->size - offset - 1));
+      _set_entry(ret, (const guint8 *)key, strlen(key),
+            (const guint8 *)value, size);
     }
+    avahi_free(key);
+    avahi_free(value);
   }
   return ret;
 }
@@ -257,7 +342,18 @@ _string_list_to_hash(AvahiStringList *list) {
 static void
 _hash_to_string_list_foreach(gpointer key, gpointer value, gpointer data) {
   AvahiStringList **list = (AvahiStringList **)data;
-  *list = avahi_string_list_add_pair(*list, key, value);
+  SalutAvahiEntryGroupServiceEntry *kentry =
+      (SalutAvahiEntryGroupServiceEntry *)key;
+  SalutAvahiEntryGroupServiceEntry *ventry =
+      (SalutAvahiEntryGroupServiceEntry *)value;
+
+  if (value != NULL) {
+    *list = avahi_string_list_add_pair_arbitrary(*list, (gchar *)kentry->value,
+         ventry->value, ventry->size);
+  } else {
+    *list = avahi_string_list_add_arbitrary(*list,
+        kentry->value, kentry->size);
+  }
 }
 
 static AvahiStringList *
@@ -270,14 +366,14 @@ _hash_to_string_list(GHashTable *table) {
 
 
 SalutAvahiEntryGroupService *
-salut_avahi_entry_group_add_service_strlist(SalutAvahiEntryGroup *group, 
-                                            const gchar *name, 
-                                            const gchar *type, 
+salut_avahi_entry_group_add_service_strlist(SalutAvahiEntryGroup *group,
+                                            const gchar *name,
+                                            const gchar *type,
                                             guint16 port,
                                             GError **error,
                                             AvahiStringList *txt) {
-  return salut_avahi_entry_group_add_service_full_strlist(group, 
-                                                          AVAHI_IF_UNSPEC, 
+  return salut_avahi_entry_group_add_service_full_strlist(group,
+                                                          AVAHI_IF_UNSPEC,
                                                           AVAHI_PROTO_UNSPEC,
                                                           0,
                                                           name, type,
@@ -286,23 +382,23 @@ salut_avahi_entry_group_add_service_strlist(SalutAvahiEntryGroup *group,
 }
 
 SalutAvahiEntryGroupService *
-salut_avahi_entry_group_add_service_full_strlist(SalutAvahiEntryGroup *group, 
+salut_avahi_entry_group_add_service_full_strlist(SalutAvahiEntryGroup *group,
                                                  AvahiIfIndex interface,
                                                  AvahiProtocol protocol,
                                                  AvahiPublishFlags flags,
-                                                 const gchar *name, 
-                                                 const gchar *type, 
-                                                 const gchar *domain, 
-                                                 const gchar *host, 
+                                                 const gchar *name,
+                                                 const gchar *type,
+                                                 const gchar *domain,
+                                                 const gchar *host,
                                                  guint16 port,
                                                  GError **error,
                                                  AvahiStringList *txt) {
-  SalutAvahiEntryGroupPrivate *priv = 
+  SalutAvahiEntryGroupPrivate *priv =
     SALUT_AVAHI_ENTRY_GROUP_GET_PRIVATE (group);
   SalutAvahiEntryGroupServicePrivate *service = NULL;
   int ret;
 
-  ret = avahi_entry_group_add_service_strlst(priv->group, 
+  ret = avahi_entry_group_add_service_strlst(priv->group,
                                              interface, protocol,
                                              flags,
                                              name, type,
@@ -311,12 +407,12 @@ salut_avahi_entry_group_add_service_full_strlist(SalutAvahiEntryGroup *group,
                                              txt);
   if (ret) {
     if (error != NULL) {
-      *error = g_error_new(SALUT_AVAHI_ERRORS, ret, 
+      *error = g_error_new(SALUT_AVAHI_ERRORS, ret,
                            "Adding service to group failed: %s",
                            avahi_strerror(ret));
     }
     goto out;
-  } 
+  }
 
   service = g_new0(SalutAvahiEntryGroupServicePrivate, 1);
   service->public.interface  = interface;
@@ -329,15 +425,16 @@ salut_avahi_entry_group_add_service_full_strlist(SalutAvahiEntryGroup *group,
   service->public.port       = port;
   service->group               = group;
   service->frozen             = FALSE;
-  service->entries            = _string_list_to_hash(txt); 
+  service->entries            = _string_list_to_hash(txt);
+  g_hash_table_insert (priv->services, group, service);
 out:
   return (SalutAvahiEntryGroupService *)service;
 }
 
 SalutAvahiEntryGroupService *
-salut_avahi_entry_group_add_service(SalutAvahiEntryGroup *group, 
-                                    const gchar *name, 
-                                    const gchar *type, 
+salut_avahi_entry_group_add_service(SalutAvahiEntryGroup *group,
+                                    const gchar *name,
+                                    const gchar *type,
                                     guint16 port,
                                     GError **error,
                                     ...) {
@@ -347,8 +444,8 @@ salut_avahi_entry_group_add_service(SalutAvahiEntryGroup *group,
   va_start(va, error);
   txt = avahi_string_list_new_va(va);
 
-  ret = salut_avahi_entry_group_add_service_full_strlist(group, 
-                                                         AVAHI_IF_UNSPEC, 
+  ret = salut_avahi_entry_group_add_service_full_strlist(group,
+                                                         AVAHI_IF_UNSPEC,
                                                          AVAHI_PROTO_UNSPEC,
                                                          0,
                                                          name, type,
@@ -360,14 +457,14 @@ salut_avahi_entry_group_add_service(SalutAvahiEntryGroup *group,
 }
 
 SalutAvahiEntryGroupService *
-salut_avahi_entry_group_add_service_full(SalutAvahiEntryGroup *group, 
+salut_avahi_entry_group_add_service_full(SalutAvahiEntryGroup *group,
                                          AvahiIfIndex interface,
                                          AvahiProtocol protocol,
                                          AvahiPublishFlags flags,
-                                         const gchar *name, 
-                                         const gchar *type, 
-                                         const gchar *domain, 
-                                         const gchar *host, 
+                                         const gchar *name,
+                                         const gchar *type,
+                                         const gchar *domain,
+                                         const gchar *host,
                                          guint16 port,
                                          GError **error,
                                          ...) {
@@ -389,17 +486,62 @@ salut_avahi_entry_group_add_service_full(SalutAvahiEntryGroup *group,
   return ret;
 }
 
-void 
+gboolean
+salut_avahi_entry_group_add_record(SalutAvahiEntryGroup *group,
+                                   AvahiPublishFlags flags,
+                                   const gchar *name,
+                                   guint16 type,
+                                   guint32 ttl,
+                                   const void *rdata,
+                                   gsize size,
+                                   GError **error) {
+ return salut_avahi_entry_group_add_record_full(group,
+     AVAHI_IF_UNSPEC, AVAHI_PROTO_UNSPEC, flags,
+     name, AVAHI_DNS_CLASS_IN, type, ttl, rdata, size, error);
+}
+
+gboolean
+salut_avahi_entry_group_add_record_full(SalutAvahiEntryGroup *group,
+                                        AvahiIfIndex interface,
+                                        AvahiProtocol protocol,
+                                        AvahiPublishFlags flags,
+                                        const gchar *name,
+                                        guint16 clazz,
+                                        guint16 type,
+                                        guint32 ttl,
+                                        const void *rdata,
+                                        gsize size,
+                                        GError **error) {
+  int ret;
+  SalutAvahiEntryGroupPrivate *priv =
+      SALUT_AVAHI_ENTRY_GROUP_GET_PRIVATE(group);
+  g_assert(group != NULL && priv->group != NULL);
+
+  ret = avahi_entry_group_add_record( priv->group, interface, protocol,
+            flags, name, clazz, type, ttl, rdata, size);
+  if (ret) {
+    if (error != NULL ) {
+      *error = g_error_new(SALUT_AVAHI_ERRORS, ret,
+                           "Setting raw record failed: %s",
+                           avahi_strerror(ret));
+    }
+    return FALSE;
+  }
+  return TRUE;
+}
+
+
+void
 salut_avahi_entry_group_service_freeze(SalutAvahiEntryGroupService *service) {
-  SalutAvahiEntryGroupServicePrivate *p = 
+  SalutAvahiEntryGroupServicePrivate *p =
     (SalutAvahiEntryGroupServicePrivate *) service;
   p->frozen = TRUE;
 }
 
 gboolean
-salut_avahi_entry_group_service_thaw(SalutAvahiEntryGroupService *service, 
+salut_avahi_entry_group_service_thaw(SalutAvahiEntryGroupService *service,
                                      GError **error) {
-  SalutAvahiEntryGroupServicePrivate *priv = 
+  SalutAvahiEntryGroupServicePrivate *priv =
     (SalutAvahiEntryGroupServicePrivate *) service;
   int ret;
   gboolean result = TRUE;
@@ -417,7 +559,7 @@ salut_avahi_entry_group_service_thaw(SalutAvahiEntryGroupService *service,
   if (ret) {
     if (error != NULL ) {
       *error = g_error_new(SALUT_AVAHI_ERRORS, ret,
-                           "Updating txt record failed: %s", 
+                           "Updating txt record failed: %s",
                            avahi_strerror(ret));
     }
     result = FALSE;
@@ -430,24 +572,48 @@ salut_avahi_entry_group_service_thaw(SalutAvahiEntryGroupService *service,
 
 gboolean
 salut_avahi_entry_group_service_set(SalutAvahiEntryGroupService *service,
-                                     const gchar *key, const gchar *value, 
+                                     const gchar *key, const gchar *value,
                                      GError **error) {
-  SalutAvahiEntryGroupServicePrivate *priv = 
+  return salut_avahi_entry_group_service_set_arbitrary(service, key,
+             (const guint8 *)value, strlen(value), error);
+
+}
+
+gboolean
+salut_avahi_entry_group_service_set_arbitrary(
+    SalutAvahiEntryGroupService *service,
+    const gchar *key, const guint8 *value, gsize size,
+    GError **error) {
+  SalutAvahiEntryGroupServicePrivate *priv =
     (SalutAvahiEntryGroupServicePrivate *) service;
 
-  g_hash_table_insert(priv->entries, g_strdup(key), g_strdup(value));
+  _set_entry(priv->entries, (const guint8 *)key, strlen(key), value, size);
 
-  if (!priv->frozen) 
+  if (!priv->frozen)
     return salut_avahi_entry_group_service_thaw(service, error);
-  else 
+  else
+    return TRUE;
+}
+
+gboolean
+salut_avahi_entry_group_service_remove_key(SalutAvahiEntryGroupService *service,
+                                           const gchar *key, GError **error) {
+  SalutAvahiEntryGroupServicePrivate *priv =
+    (SalutAvahiEntryGroupServicePrivate *) service;
+
+  g_hash_table_remove(priv->entries, key);
+
+  if (!priv->frozen)
+    return salut_avahi_entry_group_service_thaw(service, error);
+  else
     return TRUE;
 }
 
 
 gboolean
-salut_avahi_entry_group_attach(SalutAvahiEntryGroup *group, 
+salut_avahi_entry_group_attach(SalutAvahiEntryGroup *group,
                                SalutAvahiClient *client, GError **error) {
-  SalutAvahiEntryGroupPrivate *priv = 
+  SalutAvahiEntryGroupPrivate *priv =
     SALUT_AVAHI_ENTRY_GROUP_GET_PRIVATE (group);
 
   g_assert(priv->client == NULL || priv->client == client);
@@ -463,7 +629,7 @@ salut_avahi_entry_group_attach(SalutAvahiEntryGroup *group,
     if (error != NULL ) {
       int aerrno = avahi_client_errno(client->avahi_client);
       *error = g_error_new(SALUT_AVAHI_ERRORS, aerrno,
-                           "Attaching group failed: %s", 
+                           "Attaching group failed: %s",
                            avahi_strerror(aerrno));
     }
     return FALSE;
@@ -473,7 +639,7 @@ salut_avahi_entry_group_attach(SalutAvahiEntryGroup *group,
 
 gboolean
 salut_avahi_entry_group_commit(SalutAvahiEntryGroup *group, GError **error) {
-  SalutAvahiEntryGroupPrivate *priv = 
+  SalutAvahiEntryGroupPrivate *priv =
     SALUT_AVAHI_ENTRY_GROUP_GET_PRIVATE (group);
   int ret;
   ret = avahi_entry_group_commit(priv->group);
@@ -489,7 +655,7 @@ salut_avahi_entry_group_commit(SalutAvahiEntryGroup *group, GError **error) {
 
 gboolean
 salut_avahi_entry_group_reset(SalutAvahiEntryGroup *group, GError **error) {
-  SalutAvahiEntryGroupPrivate *priv = 
+  SalutAvahiEntryGroupPrivate *priv =
     SALUT_AVAHI_ENTRY_GROUP_GET_PRIVATE (group);
   int ret;
   ret = avahi_entry_group_reset(priv->group);
