@@ -20,8 +20,12 @@
 #ifndef __SALUT_SELF_H__
 #define __SALUT_SELF_H__
 
+#include "config.h"
+
 #include <glib-object.h>
 
+#include <telepathy-glib/handle-repo.h>
+
 #include "salut-avahi-client.h"
 #include "salut-presence.h"
 
@@ -39,6 +43,16 @@ struct _SalutSelf {
     gchar *name;
     SalutPresenceId status;
     gchar *status_message;
+    gchar *avatar_token;
+    guint8 *avatar;
+    gsize avatar_size;
+    gchar *jid;
+#ifdef ENABLE_OLPC
+    GArray *olpc_key;
+    gchar *olpc_cur_act;
+    TpHandle olpc_cur_act_room;
+    gchar *olpc_color;
+#endif
 };
 
 GType salut_self_get_type(void);
@@ -57,20 +71,53 @@ GType salut_self_get_type(void);
 #define SALUT_SELF_GET_CLASS(obj) \
   (G_TYPE_INSTANCE_GET_CLASS ((obj), SALUT_TYPE_SELF, SalutSelfClass))
 
-SalutSelf *
-salut_self_new(SalutAvahiClient *client,
-               gchar *first_name, gchar *last_name, 
-               gchar *jid, gchar *email);
+SalutSelf *salut_self_new (SalutAvahiClient *client,
+    TpHandleRepoIface *room_repo, const gchar *nickname,
+    const gchar *first_name, const gchar *last_name, const gchar *jid,
+    const gchar *email, const gchar *published_name,
+    const GArray *olpc_key,
+    const gchar *olpc_color);
 
 /* Start announcing our presence on the network */
-gboolean salut_self_announce(SalutSelf *self, GError **error);
+gboolean
+salut_self_announce (SalutSelf *self, gint port, GError **error);
 
-gboolean salut_self_set_presence(SalutSelf *self, 
-                                 SalutPresenceId status,
-                                 const gchar *message,
-                                 GError **error);
-const gchar 
-*salut_self_get_alias(SalutSelf *self);
+gboolean salut_self_set_presence (SalutSelf *self,
+    SalutPresenceId status, const gchar *message, GError **error);
+
+gboolean salut_self_set_avatar (SalutSelf *self, guint8 *data,
+    gsize size, GError **error);
+
+gboolean salut_self_set_alias (SalutSelf *self, const gchar *alias,
+    GError **error);
+
+const gchar *salut_self_get_alias (SalutSelf *self);
+
+#ifdef ENABLE_OLPC
+gboolean salut_self_set_olpc_properties (SalutSelf *self,
+    const GArray *key, const gchar *color, const gchar *jid, GError **error);
+
+gboolean salut_self_merge_olpc_activity_properties (SalutSelf *self,
+    TpHandle handle,
+    const gchar **color, const gchar **name, const gchar **type);
+
+gboolean salut_self_set_olpc_activity_properties (SalutSelf *self,
+    TpHandle handle,
+    const gchar *color, const gchar *name, const gchar *type,
+    GError **error);
+
+gboolean salut_self_set_olpc_activities (SalutSelf *self,
+    GHashTable *act_id_to_room, GError **error);
+
+gboolean salut_self_set_olpc_current_activity (SalutSelf *self,
+    const gchar *id, TpHandle room, GError **error);
+
+typedef void (*SalutSelfOLPCActivityFunc)
+    (const gchar *id, TpHandle handle, gpointer user_data);
+
+void salut_self_foreach_olpc_activity (SalutSelf *self,
+    SalutSelfOLPCActivityFunc foreach, gpointer user_data);
+#endif
 
 G_END_DECLS
 
